import pytest
from hypothesis import HealthCheck, assume, given, settings
from hypothesis import strategies as st
from hypothesis_jsonschema import from_schema

from jigu.core import Coin, Coins
from testtools import assert_serdes_consistent, assert_serdes_exact


@pytest.mark.sdk
class TestCoinsSerdes:
    @pytest.mark.serdes
    @pytest.mark.slow
    @settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
    @given(c=from_schema(Coins.__schema__))
    def test_serdes_consistent(self, c):
        assert_serdes_consistent(Coins, c)


@pytest.mark.sdk
class TestCoins:
    def test_constructor_copy(self):
        """sdk.Coins object should copy another sdk.Coins object when passed into constructor,
        rather than maintaining original references."""
        A = Coin("A", 1000)
        C1 = Coins([A])
        C2 = Coins(C1)

        assert C1 == C2
        assert list(C1) == list(C2)
        assert C1 is not C2
        assert C1.coins[0] is not A  # should be a new coin
        assert C2.coins[0] is not C1.coins[0]  # should be a new coin

    def test_converts_to_sorted_list(self):
        """sdk.Coins object should convert to sorted List[Coin] by alphabetic denom."""
        A = Coins([Coin("Z", 500), Coin("A", 1000)])
        B = [Coin("A", 1000), Coin("Z", 500)]
        assert list(A) == B

    def test_equality(self):
        """sdk.Coins object should be able to compare with coins and list."""
        A = Coins([Coin("A", 1000), Coin("B", 500)])
        B = Coins([Coin("B", 500), Coin("A", 1000)])
        assert A == B  # order of original input should not matter
        assert A == [Coin("A", 1000), Coin("B", 500)]
        assert [Coin("A", 1000), Coin("B", 500)] == A
        assert A == [Coin("B", 500), Coin("A", 1000)]
        assert [Coin("B", 500), Coin("A", 1000)] == A
        assert not (A == [500, 1000])
        assert not ([500, 1000] == A)

    def test_getitem(self):
        """sdk.Coins object should act like a dict, and indexable by denom as key."""

        X = Coins([Coin("A", 1000), Coin("B", 2000), Coin("C", 3000)])
        assert all(isinstance(X[i], Coin) for i in ["A", "B", "C"])
        assert X["A"] == Coin("A", 1000)
        assert X["B"] == Coin("B", 2000)
        assert X["C"] == Coin("C", 3000)
        with pytest.raises(KeyError):
            X["D"]

    def test_contains_denom(self):
        """The `in` operator in Python should allow a user to query whether a Coin of a specified denom
        exists in the sdk.Coins object."""

        haystack = Coins(
            [Coin("A", 1000), Coin("B", 500), Coin("needle", 250), Coin("C", 125)]
        )
        assert "needle" in haystack
        assert "B" in haystack
        assert "D" not in haystack

    @pytest.mark.parametrize("t", [list, Coins])
    def test_add_coins(self, t):
        """Coins can be added, which merges the collections of Coins and groups together
        the sum amount for Coins of a similar denom."""
        A = t([Coin("uluna", 2000)])
        B = t([Coin("ukrw", 1000)])
        C = t([Coin("uluna", 1000), Coin("ukrw", 250), Coin("uusd", 2000)])
        assert A + B == Coins([Coin("uluna", 2000), Coin("ukrw", 1000)])
        assert A + B + C == Coins(
            [Coin("uluna", 3000), Coin("ukrw", 1250), Coin("uusd", 2000)]
        )



from typing import Union

from jigu.client.lcd.api import ApiResponse, BaseApi, project
from jigu.core import AccAddress, Coins
from jigu.util.validation import validate_acc_address

__all__ = ["BankApi"]


class BankApi(BaseApi):
    def balance_for(self, address: AccAddress) -> Union[ApiResponse, Coins]:
        """Get's the balance of an account by its address."""
        address = validate_acc_address(address)
        res = self._api_get(f"/bank/balances/{address}")
        return project(res, Coins.deserialize(res))



import abc
import base64
import hashlib
from typing import Union

from bip32utils import BIP32_HARDEN, BIP32Key
from cached_property import cached_property

from jigu.core.auth.transaction import StdSignature, StdSignMsg, StdTx
from jigu.core.sdk.pubkey import PublicKey
from jigu.util import get_bech

BECH32_PUBKEY_DATA_PREFIX = "eb5ae98721"

__all__ = ["derive_child", "derive_root", "LUNA_COIN_TYPE", "Key"]

LUNA_COIN_TYPE = 330


def derive_root(seed: bytes) -> BIP32Key:
    return BIP32Key.fromEntropy(seed)


def derive_child(
    root: BIP32Key, account: int = 0, index: int = 0, coin_type: int = LUNA_COIN_TYPE
):
    # HD Path: 44'/330'/<acc>'/0/<idx>
    return (
        root.ChildKey(44 + BIP32_HARDEN)
        .ChildKey(coin_type + BIP32_HARDEN)
        .ChildKey(account + BIP32_HARDEN)
        .ChildKey(0)
        .ChildKey(index)
    )


class Key(object, metaclass=abc.ABCMeta):
    """Abstract interface for managing a Terra private key / public key pair."""

    @property
    @abc.abstractmethod
    def public_key(self) -> bytes:
        """Get the Key's associated public key."""
        raise NotImplementedError("Keys must implement public_key property")

    @abc.abstractmethod
    def sign(self, payload: bytes) -> bytes:
        """Signs a string payload with the private key."""
        raise NotImplementedError("Keys must implement sign()")

    @cached_property
    def base_address(self) -> bytes:
        """Get's the Key's base address for determining Bech32 acc, val addresses."""
        sha = hashlib.sha256()
        rip = hashlib.new("ripemd160")
        sha.update(self.public_key)
        rip.update(sha.digest())
        return rip.digest()

    def create_signature(self, tx: StdSignMsg) -> StdSignature:
        """Signs a transaction (`StdSignMsg` or `StdTx`), returns a signature."""
        sig_data = self.sign(tx.to_json(sort=True).strip().encode())
        pub_key = PublicKey(value=base64.b64encode(self.public_key).decode())
        return StdSignature(
            signature=base64.b64encode(sig_data).decode(), pub_key=pub_key,
        )

    def sign_tx(self, tx: Union[StdSignMsg, StdTx]) -> StdTx:
        """Creates a signature for the transaction and returns a signed transaction."""
        signature = self.create_signature(tx)
        if isinstance(tx, StdTx):  # if already signed
            tx.signatures.append(signature)
            return tx
        else:
            return StdTx(fee=tx.fee, msg=tx.msgs, signatures=[signature], memo=tx.memo)

    @cached_property
    def acc_address(self) -> str:
        """Get's the key's associated account (terra- prefixed) address."""
        return get_bech("terra", self.base_address.hex())

    @cached_property
    def acc_pubkey(self) -> str:
        """Get's the key's associated Terra public key (terrapub- prefixed)"""
        return get_bech("terrapub", BECH32_PUBKEY_DATA_PREFIX + self.public_key.hex())

    @cached_property
    def val_address(self) -> str:
        """Get's the key's associated validator operator (terravaloper- prefixed) address."""
        return get_bech("terravaloper", self.base_address.hex())

    @cached_property
    def val_pubkey(self) -> str:
        """Get's the key's associated Terra validator public key (terravaloperpub- prefixed)"""
        return get_bech(
            "terravaloperpub", BECH32_PUBKEY_DATA_PREFIX + self.public_key.hex()
        )
